# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .__version__ import __version__

__title__ = 'Reddit Terminal Viewer'
__author__ = 'Michael Lazar'
__license__ = 'The MIT License (MIT)'
__copyright__ = '(c) 2015 Michael Lazar'
